/*
 * This is Tampa Ping-Pong convers/conversd derived from the wampes
 * convers package written by Dieter Deyke <deyke@mdddhd.fc.hp.com>
 * Modifications by Fred Baumgarten <dc6iq@insu1.etec.uni-karlsruhe.de>
 *
 * Modifications by Brian A. Lantz/KO4KS <brian@lantz.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "version.h"
#include "conversd.h"
#include "host.h"
#include "user.h"
#include "convert.h"
#include "log.h"
#include "config.h"
#include "hmalloc.h"

/*---------------------------------------------------------------------------*/

void h_away_command(struct connection *cp)
{
	char *fromname, *hostname, *text;
	int channel;
	time_t time = 0;
	struct connection *p;

	fromname = getarg(0, 0);
	hostname = getargcs(0, 0);
	channel = atoi(getarg(0, 0));
	if (!user_check(cp, fromname))
		return;
	time = atol(getarg(0, 0));
	text = getarg(0, 1);
	cp->locked = 1;
	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER) {
			if (!strcasecmp(p->name, fromname) && !strcasecmp(p->host, hostname)) {
				strncpy(p->away, text, AWAYSIZE);
				p->atime = time;
			}
		}
	}
	send_awaymsg(fromname, hostname, time, text);
}

/*---------------------------------------------------------------------------*/

void h_cmsg_command(struct connection *cp)
{

	char *text;
	char *name;
	int channel;

	name = getarg(0, 0);
	channel = atoi(getarg(0, 0));
	text = getarg(0, 1);
	
	if (!user_check(cp, name))
		return;
	if (channel > MAXCHANNEL)
		return;
		
	if (*(text - 1) == ' ') {
		do {
			text--;
		} while (*(text - 1) == ' ');
	}
	if ((*text) && !((!strcmp(name, "conversd")) && strstr(text, "is calling CQ")))
		send_msg_to_channel(name, channel, text);
}

/*---------------------------------------------------------------------------*/

void h_ecmd_command(struct connection *cp)
{
	struct cmdtable *cmdp;
	char *arg, *user, *params;
	struct connection *usercp;
	int arglen;
	struct extendedcmds *eadd;
	char buffer[64];

	user = getarg(0, 0);
	arglen = strlen(arg = getarg(0, 0));
	params = getarg(0, 1);

	/* re-construct the original line to pass on to others */
	strcpy(buffer, arg);

	if (!strcasecmp(user, "conversd")) {	/* adding a new command */
		if (!ecmd_exists(buffer)) {
			eadd = hmalloc(sizeof(struct extendedcmds));
			eadd->next = ecmds;
			eadd->name = hmalloc(arglen + 1);
			strcpy(eadd->name, buffer);
			ecmds = eadd;
		}
		sprintf(cp->ibuf, "/\377\200ECMD conversd %s\n", buffer);
		h_unknown_command(cp);
		return;
	}
	/* find the actual user who sent the command */
	for (usercp = connections; usercp; usercp = usercp->next) {
		if (!strcasecmp(usercp->name, user))
			break;
	}
	if (usercp == NULLCONNECTION)
		return;

	/* now lookup the command locally */
	for (cmdp = cmdtable; cmdp->name; cmdp++)
		if (!strncmp(cmdp->name, buffer, arglen)) {
			if (cmdp->states & (1 << usercp->type)) {
				sprintf(usercp->ibuf, "%s %s\n", buffer, params);
				(void) getarg(usercp->ibuf, 0);
				(*cmdp->fnc) (usercp);
			}
			return;
		}
	/* if not found locally, then send it on to others */
	sprintf(cp->ibuf, "/\377\200ECMD %s %s %s\n", user, buffer, params);
	h_unknown_command(cp);
}

/*---------------------------------------------------------------------------*/

void h_help_command(struct connection *cp)
{
	char *user, *arg;
	char buffer[100];

	user = getarg(0, 0);
	arg = getarg(0, 0);

	/* now lookup the help locally */
	if (!dohelper(user, arg)) {
		/* if not found locally, then send it on to others */
		sprintf(cp->ibuf, "/\377\200HELP %s %s\n", user, arg);
		h_unknown_command(cp);
	} else {
		sprintf(buffer, "*** Extended Help Reply from host: %s\n", myhostname);
		send_msg_to_user("conversd", user, buffer);
		send_msg_to_user("conversd", user, "***");
	}
}

/*---------------------------------------------------------------------------*/

void h_dest_command(struct connection *cp)
{
	char *name, *rev;
	int pl, rtt;
	struct permlink *l = NULLPERMLINK;
	struct destination *dest;

	name = getargcs(0, 0);
	rtt = atoi(getarg(0, 0));
	rev = getarg(0, 0);
	if (!strcasecmp(name, myhostname))
		return;
	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		l = permarray[pl];
		if (l && l->connection && (l->connection == cp))
			break;
	}
	for (dest = destinations; dest; dest = dest->next) {
		if (!strcasecmp(dest->name, name))
			break;
	}
	if (l && dest_check(cp, dest))
		update_destinations(l, name, rtt, rev);
}

/*---------------------------------------------------------------------------*/

void h_host_command(struct connection *cp)
{
	char *name;
	char *rev;
	char *features;
	char buffer[2048];
	struct connection *p;
	struct destination *d;
	struct permlink *pp;
	struct channel *ch;
	struct clist *cl;
	int pl;
	char flags[16];

	name = getargcs(0, 0);
	rev = getarg(0, 0);
	features = getargcs(0, 0);
	if (!*name)
		return;
	
	if (!cp->hostallowed) {
		log(L_AUTH, "Refused HOST connection from %s [%s]", name, inet_ntoa(cp->sin.sin_addr));
		return;
	}
	
	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		pp = permarray[pl];
		if (pp && pp->connection && (pp->connection != cp) && !strcasecmp(pp->name, name)) {
			bye_command2(pp->connection, "link reorg");
		}
		update_destinations(pp, name, 0, rev);
	}
	if (cp->type != CT_UNKNOWN)
		return;
	cp->type = CT_HOST;
	strncpy(cp->rev, rev, NAMESIZE);
	strncpy(cp->name, name, NAMESIZE);
	route_cleanup(cp);
	cp->charset_in = cp->charset_out = dumb;
	if (strchr(features, 'a'))
		cp->features |= FEATURE_AWAY;
	if (strchr(features, 'd'))
		cp->features |= FEATURE_FWD;
	if (strchr(features, 'm'))
		cp->features |= FEATURE_MODES;
	if (strchr(features, 'p'))
		cp->features |= FEATURE_LINK;
	if (strchr(features, 'u'))
		cp->features |= FEATURE_UDAT;
	if (strchr(features, 'n'))
		cp->features |= FEATURE_NICK;
/* #define DEBUG */
#ifdef DEBUG
	printf("Features '%s': %s%s%s%s%s%s\n", name,
	       (cp->features & FEATURE_AWAY) ? "a" : "",
	       (cp->features & FEATURE_FWD) ? "d" : "",
	       (cp->features & FEATURE_MODES) ? "m" : "",
	       (cp->features & FEATURE_LINK) ? "p" : "",
	       (cp->features & FEATURE_UDAT) ? "u" : "",
	       (cp->features & FEATURE_NICK) ? "n" : "");
#endif

	if (strchr(features, 'u'))
		cp->amprnet = 1;
	update_permlinks(name, cp, 0);
	sprintf(buffer, "/\377\200HOST %s %s %s\n", myhostname, SOFT_REL8, myfeatures);
	linklog(cp, L_SENT, "%s", &buffer[3]);
	appendstring(cp, buffer);
	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER) {
			if (!p->via) {
				for (cl = p->chan_list; cl; cl = cl->next) {
					if (cp->amprnet) {
						sprintf(buffer, "/\377\200%s %s %s %ld -1 %d %s\n", (p->observer) ? "OBSV" : "USER", p->name, p->host, (long)cl->time, cl->channel, p->pers);
						linklog(cp, L_SENT, "%s", &buffer[3]);
						appendstring(cp, buffer);
					} else {
						sprintf(buffer, "/\377\200%s %s %s %ld -1 %d\n", (p->observer) ? "OBSV" : "USER", p->name, p->host, (long)cl->time, cl->channel);
						linklog(cp, L_SENT, "%s", &buffer[3]);
						appendstring(cp, buffer);
						if (!p->observer && strcmp(p->name, "@")) {
							sprintf(buffer, "/\377\200UDAT %s %s %s\n", p->name, p->host, p->pers);
							linklog(cp, L_SENT, "%s", &buffer[3]);
							appendstring(cp, buffer);
						}
					}
					if (cp->features & FEATURE_NICK) {
						sprintf(buffer, "/\377\200UADD %s %s %s %d %s\n", p->name, p->host, p->nickname, -1, p->pers ? p->pers : "~~");
						linklog(cp, L_SENT, "%s", &buffer[3]);
						appendstring(cp, buffer);
					}
					if (*p->away) {
						if (cp->oldaway)
							sprintf(buffer, "/\377\200AWAY %s %s 69 %ld %s\n", p->name, p->host, (long)p->atime, p->away);
						else
							sprintf(buffer, "/\377\200AWAY %s %s %ld %s\n", p->name, p->host, (long)p->atime, p->away);
						linklog(cp, L_SENT, "%s", &buffer[3]);
						appendstring(cp, buffer);
					}
					if (cl->channelop) {
						sprintf(buffer, "/\377\200OPER conversd %d %s\n", cl->channel, p->name);
						linklog(cp, L_SENT, "%s", &buffer[3]);
						appendstring(cp, buffer);
					}
					if (p->operator) {
						sprintf(buffer, "/\377\200OPER conversd -1 %s\n", p->name);
						linklog(cp, L_SENT, "%s", &buffer[3]);
						appendstring(cp, buffer);
					}
				}
			} else {
				if (cp->amprnet) {
					sprintf(buffer, "/\377\200%s %s %s %ld -1 %d %s\n", (p->observer) ? "OBSV" : "USER", p->name, p->host, (long)p->time, p->channel, p->pers);
					linklog(cp, L_SENT, "%s", &buffer[3]);
					appendstring(cp, buffer);
				} else {
					sprintf(buffer, "/\377\200%s %s %s %ld -1 %d\n", (p->observer) ? "OBSV" : "USER", p->name, p->host, (long)p->time, p->channel);
					linklog(cp, L_SENT, "%s", &buffer[3]);
					appendstring(cp, buffer);
					if (!p->observer) {
						sprintf(buffer, "/\377\200UDAT %s %s %s\n", p->name, p->host, p->pers);
						linklog(cp, L_SENT, "%s", &buffer[3]);
						appendstring(cp, buffer);
					}
				}
				if (*p->away) {
					sprintf(buffer, "/\377\200AWAY %s %s %d %ld %s\n", p->name, p->host, p->channel, (long)p->atime, p->away);
					linklog(cp, L_SENT, "%s", &buffer[3]);
					appendstring(cp, buffer);
				}
				if (p->channelop) {
					sprintf(buffer, "/\377\200OPER conversd %d %s\n", p->channel, p->name);
					linklog(cp, L_SENT, "%s", &buffer[3]);
					appendstring(cp, buffer);
				}
				if (p->operator) {
					sprintf(buffer, "/\377\200OPER conversd -1 %s\n", p->name);
					linklog(cp, L_SENT, "%s", &buffer[3]);
					appendstring(cp, buffer);
				}
			}
		}
	}
	for (ch = channels; ch; ch = ch->next) {
		if (*ch->topic) {
			sprintf(buffer, "/\377\200TOPI conversd %s %ld %d %s\n", myhostname, (long)ch->time, ch->chan, ch->topic);
			linklog(cp, L_SENT, "%s", &buffer[3]);
			appendstring(cp, buffer);
		}
		flags[0] = '\0';
		if (ch->flags & M_CHAN_S) {
			strcat(flags, "s");
		}
		if (ch->flags & M_CHAN_P) {
			strcat(flags, "p");
		}
		if (ch->flags & M_CHAN_T) {
			strcat(flags, "t");
		}
		if (ch->flags & M_CHAN_I) {
			strcat(flags, "i");
		}
		if (ch->flags & M_CHAN_M) {
			strcat(flags, "m");
		}
		if (ch->flags & M_CHAN_L) {
			strcat(flags, "l");
		}
		sprintf(buffer, "/\377\200MODE %d -sptiml+%s\n", ch->chan, flags);
		linklog(cp, L_SENT, "%s", &buffer[3]);
		appendstring(cp, buffer);
	}
	for (d = destinations; d; d = d->next) {
		if (d->rtt) {
			sprintf(buffer, "/\377\200DEST %s %ld %s\n", d->name, d->rtt + 99, d->rev);
			linklog(cp, L_SENT, "%s", &buffer[3]);
			appendstring(cp, buffer);
		}
	}
}

/*---------------------------------------------------------------------------*/

/* this command simply passes on any host requests that we don't understand,
   since our neighbors MIGHT understand them */

void h_unknown_command(struct connection *cp)
{
	struct connection *p;
	
	for (p = connections; p; p = p->next) {
		if (p->type == CT_HOST && p != cp) {
			linklog(p, L_SENT, "%s", &cp->ibuf[3]);
			appendstring(p, cp->ibuf);
		}
	}
}


/*---------------------------------------------------------------------------*/

void h_invi_command(struct connection *cp)
{

	char *fromname, *toname;
	int channel;

	fromname = getarg(0, 0);
	toname = getarg(0, 0);
	channel = atoi(getarg(0, 0));
	if (!user_check(cp, fromname))
		return;
	send_invite_msg(fromname, toname, channel);
}

/*---------------------------------------------------------------------------*/

void h_link_command(struct connection *cp)
{
	char *fromname, *host;
	char buffer[100];

	fromname = getargcs(0, 0);
	host = getargcs(0, 0);
	if ((strlen(fromname) > NAMESIZE) || (strlen(host) > NAMESIZE))
		return;
	if (!strcasecmp(host, myhostname)) {
		if (!user_check(cp, fromname))
			return;
		clear_locks();
		sprintf(buffer, "*** Links at %s", myhostname);
		send_msg_to_user("conversd", fromname, buffer);
		display_links_command(NULLCONNECTION, fromname);
		return;
	}
	/* else, pass it on */
	sprintf(buffer, "/\377\200LINK %s %s\n", fromname, host);
	strcpy(cp->ibuf, buffer);
	h_unknown_command(cp);
}

/*---------------------------------------------------------------------------*/

void h_oper_command(struct connection *cp)
{
	char *toname, *fromname;
	int channel;
	struct connection *p;
	struct clist *cl;

	fromname = getarg(0, 0);
	channel = atoi(getarg(0, 0));
	toname = getarg(0, 0);
	cp->locked = 1;

	for (p = connections; p; p = p->next) {
		if (channel == -1) {
			if ((!strcasecmp(p->name, toname)) && (p->type == CT_USER)) {
				p->operator = 1;
				if (!user_check(cp, fromname))
					send_opermsg(toname, p->host, fromname, -1);
			}
		} else {
			if (!strcasecmp(p->name, toname)) {
				if (p->channel == channel) {
					p->channelop = 1;
				}
				for (cl = p->chan_list; cl; cl = cl->next) {
					if (cl->channel == p->channel)
						cl->channelop = 1;
				}
				send_opermsg(toname, p->host, cp->name, p->channel);
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

void h_ping_command(struct connection *cp)
{
	char buffer[2048];
	int pl;

	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		if (permarray[pl] && (permarray[pl]->connection == cp)) {
			sprintf(buffer, "/\377\200PONG %ld\n", permarray[pl]->txtime);
			linklog(cp, L_SENT, "%s", &buffer[3]);
			appendstring(cp, buffer);
		}
	}
}

/*---------------------------------------------------------------------------*/

void h_pong_command(struct connection *cp)
{
	struct permlink *l;
	int pl;

	l = NULLPERMLINK;
	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		if (permarray[pl] && (permarray[pl]->connection == cp)) {
			l = permarray[pl];
			l->rxtime = atol(getarg(0, 0));
			l->txtime = max(currtime - l->testwaittime, 1);
			if (l->rxtime == 0)
				l->rxtime = l->txtime;
			if ((abs(l->rxtime) > 20001) || (abs(l->txtime) > 20001)) {
				l->rxtime = 0;
				l->txtime = 0;
			}
			l->testnexttime = l->testwaittime + max(min(60 * (l->txtime), 7200), 120);
			/* I hacked this because of it's nasty behave - rewrite of the whole stuff is in progress */
			update_destinations(l, l->name,
					    (l->rxtime == -1) ? l->txtime : (l->rxtime + l->txtime) / 2L,
					    cp->rev);
			break;
		}
	}
}

/*---------------------------------------------------------------------------*/

void h_rout_command(struct connection *cp)
{
	char *dest, *user;
	struct destination *d;
	char buffer[2048];
	int ttl;
	int goodlink;

	dest = getargcs(0, 0);
	user = getarg(0, 0);
	ttl = atoi(getarg(0, 0));
	if (!user_check(cp, user))
		return;
	clear_locks();
	for (d = destinations; d; d = d->next) {
		goodlink = (d->link && d->link->name && *d->link->name);
		if (!strcasecmp(d->name, dest)) {
			sprintf(buffer, "*** route: %s -> %s -> %s (%ld)", myhostname,
				(goodlink) ? d->link->name : "(dead link)", dest, d->rtt);
			send_msg_to_user("conversd", user, buffer);
			if (ttl && goodlink && strcasecmp(d->link->name, dest)) {
				ttl--;
				sprintf(buffer, "/\377\200ROUT %s %s %d\n", dest, user, ttl);
				linklog(cp, L_SENT, "%s", &buffer[3]);
				appendstring(d->link->connection, buffer);
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

void h_sysi_command(struct connection *cp)
{
	char *user, *host;
	char buffer[256];

	user = getargcs(0, 0);
	host = getargcs(0, 0);
	cp->locked = 0;
	if (!user_check(cp, user))
		return;

	if (!strcasecmp(host, myhostname) || !strcasecmp(host, "all")) {	/* if for us */
		sprintf(buffer, "*** System Information for: %s - email to %s\n*** ", myhostname, (myemailaddr) ? myemailaddr : "*** unknown ***");
		send_msg_to_user("conversd", user, buffer);
		clear_locks();
		sprintf(buffer, "*** %s: %s\n", myhostname, (mysysinfo) ? mysysinfo : "No additional information available");
		send_msg_to_user("conversd", user, buffer);
		if (strcasecmp(host, "all"))
			return;
	}
	/* else, pass it on */
	sprintf(buffer, "/\377\200SYSI %s %s\n", user, host);
	strcpy(cp->ibuf, buffer);
	h_unknown_command(cp);
}

/*---------------------------------------------------------------------------*/

void h_topi_command(struct connection *cp)
{
	char *fromname, *hostname, *text;
	int channel;
	time_t time;
	
	fromname = getarg(0, 0);
	hostname = getargcs(0, 0);
	time = atol(getarg(0, 0));
	channel = atoi(getarg(0, 0));
	text = getarg(0, 1);
	
	if (!strcasecmp(fromname, "zeus")) {
		log(L_ERR, "Ancient gods detected in the direction of %s, ignoring their madness.", cp->name);
		return;
	}
	
	if ((!strcasecmp(fromname, "conversd")) && (strstr(text, "may not be longer than"))) {
		log(L_ERR, "Echoes of ancient gods heard in the direction of %s, ignoring it's foolishness.", cp->name);
		return;
	}
	
	if (!user_check(cp, fromname))
		return;
	cp->locked = 1;
	send_topic(fromname, hostname, time, channel, text);
}

/*---------------------------------------------------------------------------*/

void h_udat_command(struct connection *cp)
{
	char *fromname, *hostname, *text;
	struct connection *p;

	fromname = getarg(0, 0);
	hostname = getargcs(0, 0);
	text = getarg(0, 1);
	if (!user_check(cp, fromname))
		return;
	cp->locked = 1;
	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER && !strcasecmp(p->name, fromname) &&
		    !strcasecmp(p->host, hostname) && strcasecmp(p->pers, text)) {
			if (p) {
				strncpy(p->pers, text, PERSSIZE);
			}
			send_persmsg(fromname, hostname, p->channel, text, currtime);
		}
	}
}

/*---------------------------------------------------------------------------*/

void h_umsg_command(struct connection *cp)
{
	char *fromname, *toname, *text;

	fromname = getarg(0, 0);
	toname = getarg(0, 0);
	text = getarg(0, 1);
	if (!user_check(cp, fromname))
		return;
	if ((!strcmp(fromname, "conversd")) && (strstr(text, "calling CQ")))
		return;
	if (*text)
		send_msg_to_user(fromname, toname, text);
}

/*---------------------------------------------------------------------------*/

void h_user_command(struct connection *cp)
{
	char *host;
	char *name;
	char *pers;
	int newchannel = 0;
	int oldchannel;
	struct connection *p;
	struct channel *ch;
	int users_left;
	time_t time;

	pers = name = getarg(0, 0);
	if (!*pers)
		return;		/* invalid user name */
	if (!user_check(cp, name))
		return;
	while (*pers) {
		if (!isgraph(*pers))
			return;	/* invalid user name */
		pers++;
	}
	pers = host = getargcs(0, 0);
	if (!*pers)
		return;		/* invalid host name */

	if (!strcmp(pers, "-1"))
		return;		/* invalid host name */

	while (*pers) {
		if (!isgraph(*pers))
			return;	/* invalid host name */
		if (!isdigit(*pers))
			newchannel++;
		pers++;
	}
	if (!newchannel)
		return;		/* invalid host name - all digits */
	time = atol(getarg(0, 0));
	if (time == 0)
		time = currtime;
	oldchannel = atoi(getarg(0, 0));
	newchannel = atoi(getarg(0, 0));
	pers = getarg(0, 1);
	
	if (oldchannel > MAXCHANNEL)
		oldchannel = -1;
	if (newchannel > MAXCHANNEL)
		newchannel = -1;
	
	if (oldchannel == -1 && newchannel == -1)
		return;
	
	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER) {
			/* new 920705 dl9sau */
			/* If Neighbour2 registers a user on HostX, while someone has already
			 * been registered for HostX via Neighbour1, then we definitely have
			 * a loop !  We send a loop detect message and then close the link:
			 * /..LOOP <Chostname> <myneighbour> <host>
			 *
			 * The LOOP PREVENTION CODE detects ONLY a loop if it starts at this
			 * host. That's, why I suggest this code to be implemented in every
			 * conversd implementation.
			 */
			if (oldchannel < 0 && p->via != cp && !strcasecmp(p->host, host)) {
				char buf[200];
				sprintf(buf, "/\377\200LOOP %s %s %s\n", myhostname, host, p->via ? p->via->name : myhostname);
				linklog(cp, L_SENT, "%s", &buf[3]);
				appendstring(cp, buf);
				log(L_R_ERR, "conversd sent: LOOP %s", host);
				bye_command(cp);
				return;
			}
			if ((p->via == cp) &&
			    !strcasecmp(p->name, name) &&
			    !strcasecmp(p->host, host) &&
			    p->channel == oldchannel)
				break;
		}
	}
	if (!p) {
		p = (struct connection *)hcalloc(1, sizeof(struct connection));
		p->type = CT_USER;
		strncpy(p->name, name, NAMESIZE);
		strncpy(p->nickname, name, NAMESIZE);
		strncpy(p->host, host, NAMESIZE);
		p->via = cp;
		p->channel = oldchannel;
		p->fd = -1;
		p->query[0] = '\0';
		p->notify[0] = '\0';
		p->next = connections;
		connections = p;
		log(L_NUSER, "New user %s@%s", name, host);
	}
	p->time = time;
	p->mtime = time;
	p->observer = (tolower(cp->ibuf[3]) == 'o');
	if (pers && *pers) {
		strncpy(p->pers, pers, PERSSIZE);
		if (!strcasecmp(pers, OBSERVERID))
			p->observer = 1;
	}
	if ((p->channel = newchannel) < 0) {
		p->type = CT_CLOSED;
	}
	users_left = count_user(oldchannel);
	if (!users_left)
		destroy_channel(oldchannel);
	for (ch = channels; ch; ch = ch->next) {
		if (ch->chan == newchannel)
			break;
	}
	if (!ch && newchannel != -1)
		new_channel(newchannel);
	send_user_change_msg(name, host, oldchannel, newchannel, pers, time, p->observer);
}
